/**
 * Abstract class that keeps track of the intersection that occurred
 *   Things like normal, etc should be placed in here. For now it's a
 *   dummy class, but soon it will be used heavily.
 *
 * Most importantly, we'll keep to instantiate only a few of these (aka
 *   one per thread) and keep updating information within them.   
 *
 * @author:	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date:	09.04.11
 */
#ifndef INTERSECTION_RECORD_H
#define INTERSECTION_RECORD_H


#include "Vector.h"
#include "Constants.h"
#include "MatPrimitive.h"


/// definition of intersection record
class IntersectionRecord {

	public:
#ifdef DO_MULTIOBJECT
		/**
		 * Constructor
		 * @param hitPt   - hit point for this intersection record
		 * @param geomhit - pointer to the geometry that we intersected with
		 * @param oType   - type of object we intersected with
		 */
		IntersectionRecord( const float &hitPt=MYRTC::RAY_INF, const void* const geomHit=0,
							const MYRTC::OBJ_TYPE &oType=MYRTC::OT_UNKNOWN, const MatPrimitive* const geomMat=0 ) :
							m_fHitRayT(hitPt), m_oType(oType), m_iGeomHitAddr(0), m_iGeomMatID(0) {
			 m_pGeomHit = (void*)( geomHit );
			 m_pGeomMat = (MatPrimitive*)( geomMat );
		}


		// public for speed
		float m_fHitRayT;			/// t value of the ray where intersection happened
		int m_iGeomHitAddr;			/// TRaX material address for hit geometry
		int m_iGeomMatID;			/// TRaX material id for hit geometry
		void *m_pGeomHit;			/// pointer to the geometry that we hit
		MYRTC::OBJ_TYPE m_oType;	/// object type of the intersected geometry
		MatPrimitive *m_pGeomMat;	/// material pointer from the geometry
//		float m_fRayMatEta;			/// current eta for the material we're going through right now
#else
		/**
		 * Constructor
		 * @param hitPt   - hit point for this intersection record
		 * @param geomhit - pointer to the geometry that we intersected with
		 * @param oType   - type of object we intersected with
		 */
		IntersectionRecord( const float &hitPt=MYRTC::RAY_INF ) :
							m_fHitRayT(hitPt), m_iGeomHitAddr(0), m_iGeomMatID(0) {
		}

		// public for speed
		float m_fHitRayT;			/// t value of the ray where intersection happened
		int m_iGeomHitAddr;			/// TRaX material address for hit geometry
		int m_iGeomMatID;			/// TRaX material id for hit geometry
#endif
};

#endif // INTERSECTION_RECORD_H
