/**
 * Implementation of the box class
 *
 * @author	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date	09.25.11
 */
#include "gpBox.h"
#include "Constants.h"


/// Intersects a ray with this box and returns the near/far intersection points
bool gpBox::intersectDetailed( float &t0out, float &t1out, const RayPrimitive &ray ) const {

	// do what was done in "An Efficient and Robust Ray-Box Intersection Algorithm" 
	//		by  Amy Williams, Steve Barrus, R. Keith Morley, and Peter Shirley
	float tMin, tMax, tMinY, tMaxY, tMinZ, tMaxZ;

	tMin  = ( pts[ray.m_iDirSignX].x   - ray.m_vOrigin.x ) * ray.m_vDirInv.x;
	tMax  = ( pts[1-ray.m_iDirSignX].x - ray.m_vOrigin.x ) * ray.m_vDirInv.x;
	tMinY = ( pts[ray.m_iDirSignY].y   - ray.m_vOrigin.y ) * ray.m_vDirInv.y;
	tMaxY = ( pts[1-ray.m_iDirSignY].y - ray.m_vOrigin.y ) * ray.m_vDirInv.y;
	if( (tMin > tMaxY) || (tMinY > tMax ) )
		return false;
	if( tMinY > tMin )
		tMin = tMinY;
	if( tMaxY < tMax )
		tMax = tMaxY;

	tMinZ = ( pts[ray.m_iDirSignZ].z   - ray.m_vOrigin.z ) * ray.m_vDirInv.z;
	tMaxZ = ( pts[1-ray.m_iDirSignZ].z - ray.m_vOrigin.z ) * ray.m_vDirInv.z;
	if( (tMin > tMaxZ) || (tMinZ > tMax) )
		return false;
	if( tMinZ > tMin )
		tMin = tMinZ;
	if( tMaxZ < tMax )
		tMax = tMaxZ;

	// set min/max values
	t0out = tMin;
	t1out = tMax;
	return( (t0out < ray.m_fMaxT) && (t1out > ray.m_fMinT) );
}
