/**
 * This is the main file pulling together all the bits necessary for 
 *   CS 6965 Project 4 - using BVH
 *
 * This is the extension of no inheritance project 2
 *
 * @author:	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date:	10.06.11
 */

//#define DO_REFLECTIONS 1		// flag whether to compile with reflection-enabling code bits (decreases speed)
//#define DO_MULTIOBJECT 1		// flag whether to keep track of multiple objects (decreases speed)
#include "project4.h"

/**
 * Which scene do we compile to run? -> soon i'll do a generate scene file or something...
 * 0 - required image
 * 1 - meant for cornell box scene with two spheres (reflective/refractive) and procedural square texture
 * 2 - cornell box with 2 balls (no scene)
 * 3 - reproduction of TRaX cornell box scene
 * 4 - crazy scene with refractions and reflections (loaded triangles)
 */
#define TRAX_SCENE 0
#define SCENE TRAX_SCENE

#define BVH_TRAVERSAL_STACK_SIZE 32		// size of the stack used to queue up bvh boxes to go through
#define DO_BVH 1						// flag whether to do BVH traversal OR all triangles
#define SUPERSAMPLE 0					// flag whether to supersample the image pixels


// ray offsets for scenes to account for different scales
#if SCENE==1
	#define RAY_OFFSET MYRTC::RAY_INTER_STEP
#else
	#define RAY_OFFSET MYRTC::RAY_INTER_STEP
#endif



// procedural tile "material" settings
#if SCENE==4
	// scene: \test_models\testScene_p4.obj
	// ids of things:
	// 0 - buddha
	// 1 - teapot
	// 2 - tiled floor
	// 3 - back, ceiling
	// 4 - light
	// 5 - right wall
	// 6 - dragon
	// 7 - box
	// 8 - bunny
	// 9 - left wall
	const Color evenClr( 0.87f, 0.95f, 1.f );//0.7f, 0.7f, 0.3f );
	const Color oddClr( 0.8f, 0.75f, 0.75f );//0.698f, 0.698f, 0.698f );//3f, 0.7f, 0.3f );
	const float tileScale = 6.f;
	const int matIDToTile = 2;
	const int matIDToReflect  = 1;
	const int matIDToReflect2 = 7;
	const int matIDToRefract  = 6;
	const float refrEta = MYRTC::MATETA_WATER;
#elif SCENE==3
	const Color evenClr( 0.87f, 0.95f, 1.f );//0.7f, 0.7f, 0.3f );
	const Color oddClr( 0.8f, 0.75f, 0.75f );//0.698f, 0.698f, 0.698f );//3f, 0.7f, 0.3f );
	const float tileScale = 6.f;
	const int matIDToTile = 1;
#else
	const Color evenClr( 0.87f, 0.95f, 1.f );//0.7f, 0.7f, 0.3f );
	const Color oddClr( 0.8f, 0.75f, 0.75f );//0.698f, 0.698f, 0.698f );//3f, 0.7f, 0.3f );
	const float tileScale = 6.f;
	const int matIDToTile = 1;
	const int matIDToReflect = -2;//2;
	const int matIDToRefract = -2;//3;
	const float refrEta = MYRTC::MATETA_WATER;
#endif


int main() {
	trax_setup();

	if( trax_start_render_threads( &renderMe, 4 ) == false )
		renderMe(0);

	trax_cleanup();
	return 0;
}



inline void * renderMe( void * ) {
/// main code to do the entire raytracing image :D
//int main() {

	// Setup code. On the CPP target sets up global memory.
//	trax_setup();

	// Load the pointer to the frame buffer in global memory
	int framebuffer = GetFrameBuffer();
	int widthINT	= loadi( 0, 1 );
	float widthInv	= loadf( 0, 2 );
	float width		= loadf( 0, 3 );
	int heightINT	= loadi( 0, 4 );
	float heightInv	= loadf( 0, 5 );
	float height	= loadf( 0, 6 );


	// set up Required scenes
#if SCENE==TRAX_SCENE || SCENE==1 || SCENE==4		// scene 8 = required image

	// set up things we know in advance, without loading scene file
	const Color ltAmb( 0.65f, 0.65f, 0.65f );
	const float kDiff = 0.7f;
	const float kAmb  = 0.3f;

	// load BVH address
	const int sAddrBVH = loadi( 0, 8 );

	// load 1st material (to try)
	const int sAddrMat = loadi( 0, 9 );

	// load camera -> computes aspect ration and such on its own
	myCamera cam( loadi(0, 10) );

	// load background color
	const Color backgroundC( loadi(0, 11) );

	// load point light
	ltPoint ptLt( loadi(0, 12), Color(1.f,1.f,1.f) );

	// load triangle information
	const int sAddrTri = loadi( 0, 28 );
	const int numTri   = loadi( 0, 29 );

	// add spheres and such
#if SCENE==1

	// scene: \test_models\cornell\CornellBoxMine1.obj

	const int numSph = 3;
//	gpSphere sph1( 260.f,140.f, 250.f, 80.f );
	//gpSphere sph1( 0.45f, -0.07f, 0.12f, 0.07f );
//	gpSphere sph1( 0.25f, -0.27f, 0.15f, 0.17f );
	//gpSphere sph1( 0.65f, -0.27f, 0.15f, 0.35f );
	gpSphere sph1( 0.35f, -0.475f, 0.2f, 0.15f );
	//gpSphere sph1( 450.f, 90.f, 75.f, 75.f );
	//gpSphere sph2( 190.f, 140.f, 216.f, 50.f );
//	gpSphere sph2( 0.f, 559.2f, 530.f, 150.f );
	//gpSphere sph2( 0.f, -0.5487f, 0.5592f, 0.175f );
	gpSphere sph2( 0.f, -1.0974f, 1.1184f, 0.35f );
	gpSphere sph3( 0.95f, -0.15f, 0.2f, 0.15f );
	//MatPrimitive mat1( Color( 1.f, 1.f, 1.f ), 0.f, 0.f, MYRTC::MATETA_AIR, true, true );
	MatPrimitive mat1( Color( 1.f, 1.f, 1.f ), 0.f, 0.f, 1.5f, true, true );//MYRTC::MATETA_PLEXIGLASS, true, true );
	MatPrimitive mat2( Color( 0.8f, 0.8f, 0.8f ), 0.f, 0.f, 1.0f, true, false );
	sph1.m_pMat = &mat1;
	sph2.m_pMat = &mat2;
	sph3.m_pMat = &mat1;
	gpSphere* objSph[] = { &sph1, &sph2, &sph3 };

	const int numTriO = 0;
	gpTriangle* objTri[] = { 0 };

	const int numPtLights = 1;
	ltPoint* ptLts[] = { &ptLt };

	const int maxRecursionDepth = 6;

#elif SCENE==4

	// scene: \test_models\testScene_p4.obj

	const int numPtLights = 2;
	//ltPoint ptLeft( Vector( 0.6f, -0.3f, 0.f ), Color( 0.2f, 0.2f, 0.2f ) );
	ltPoint ptLeft( Vector( 0.6f, -0.4f, 0.f ), Color( 0.35f, 0.35f, 0.4f ) );
	ltPoint ptRight( Vector( 0.831f, 0.5f, 0.3f ), Color( 0.8f, 0.5f, 0.5f ) );
	ltPoint* ptLts[] = { &ptLt, &ptLeft, &ptRight };

	const int numSph = 0;
	gpSphere* objSph[] = { 0 };

	const int numTriO = 0;
	gpTriangle* objTri[] = { 0 };

	const int maxRecursionDepth = 10;
#endif

#elif SCENE==2		// scene 2 = cornell box with refractive/reflective spheres
	const Color backgroundC( 0.3137f, 0.3529f, 0.3529f );

	MatPrimitive matSphL( Color( 1.0f, 1.0f, 1.0f ), 0.0f, 0.0f, MYRTC::MATETA_AIR, true, false );
	MatPrimitive matSphR( Color( 1.0f, 1.0f, 1.0f ), 0.0f, 0.0f, MYRTC::MATETA_PLEXIGLASS, true, true );
	MatPrimitive matBoxL( Color( 0.5f, 0.5f, 0.8f ), 0.9f, 0.1f );		// box colors from UCSD
	MatPrimitive matBoxR( Color( 0.8f, 0.5f, 0.5f ), 0.9f, 0.1f );
	MatPrimitive matBoxT( Color( 0.7f, 0.7f, 0.7f ), 0.9f, 0.1f );

	gpSphere sphL( -0.4f, -0.65f, -0.4f, 0.35f );
	gpSphere sphR(  0.4f, -0.65f,  0.1f, 0.35f );
	sphL.m_pMat = &matSphL;
	sphR.m_pMat = &matSphR;

	// construct left wall
	gpTriangle l1( Vector(-1.f,-1.f,1.f), Vector(-1.f,-1.f,-1.f), Vector(-1.f,1.f,-1.f) );
	gpTriangle l2( Vector(-1.f,1.f,-1.f), Vector(-1.f,1.f,1.f), Vector(-1.f,-1.f,1.f) );
	l1.m_pMat = &matBoxL;
	l2.m_pMat = &matBoxL;

	// right wall
	gpTriangle r1( Vector(1.f,1.f,-1.f), Vector(1.f,-1.f,-1.f), Vector(1.f,-1.f,1.f) );
	gpTriangle r2( Vector(1.f,-1.f,1.f), Vector(1.f,1.f,1.f), Vector(1.f,1.f,-1.f) );
	r1.m_pMat = &matBoxR;
	r2.m_pMat = &matBoxR;

	// top wall
	gpTriangle t1( Vector(1.f,1.f,1.f), Vector(-1.f,1.f,1.f), Vector(-1.f,1.f,-1.f) );
	gpTriangle t2( Vector(-1.f,1.f,-1.f), Vector(1.f,1.f,-1.f), Vector(1.f,1.f,1.f) );
	t1.m_pMat = &matBoxT;
	t2.m_pMat = &matBoxT;

	// back wall
	gpTriangle bk1( Vector(-1.f,-1.f,-1.f), Vector(1.f,-1.f,-1.f), Vector(1.f,1.f,-1.f) );
	gpTriangle bk2(  Vector(1.f,1.f,-1.f), Vector(-1.f,1.f,-1.f), Vector(-1.f,-1.f,-1.f) );
	bk1.m_pMat = &matBoxT;
	bk2.m_pMat = &matBoxT;

	// bottom wall
	gpTriangle bt1( Vector(-1.f,-1.f,-1.f), Vector(-1.f,-1.f,1.f), Vector(1.f,-1.f,1.f) );
	gpTriangle bt2( Vector(1.f,-1.f,1.f), Vector(1.f,-1.f,-1.f), Vector(-1.f,-1.f,-1.f) );
	bt1.m_pMat = &matBoxT;
	bt2.m_pMat = &matBoxT;

	MatPrimitive matLt( Color( 20.f, 20.f, 20.f ), 0.f, 1.f );
	gpTriangle lt1( Vector(-0.45f,0.9999f,-0.3f), Vector(0.45f,0.9999f,-0.3f), Vector(0.45f,0.9999f,0.3f) );
	gpTriangle lt2( Vector(0.45f,0.9999f,0.3f), Vector(-0.45f,0.9999f,0.3f), Vector(-0.45f,0.9999f,-0.3f) );
	lt1.m_pMat = &matLt;
	lt2.m_pMat = &matLt;

	const int numSph = 2;
	gpSphere* objSph[] = { &sphL, &sphR };

	const int numTriO = 12;
	gpTriangle* objTri[] = { &l1, &l2, &r1, &r2, &t1, &t2, &bk1, &bk2, &bt1, &bt2, &lt1, &lt2 };

	myCamera cam( Vector(0.f,0.f,3.8f), Vector(0.f,0.f,0.f), Vector(0.f,1.f,0.f),
					width, height, MYRTC::CAM_FOV_193_DEG );

	const int numPtLights = 1;
	Color ltAmb( 1.f, 1.f, 1.f );
	ltPoint ltPt1( Vector(0.f, 0.9f, 0.f), Color(1.f, 1.f, 1.f) );
	ltPoint* ptLts[] = { &ltPt1 };

	const int maxRecursionDepth = 10;

	const float kDiff  = 0.7f;
	const float kAmb   = 0.3f;
	const int sAddrBVH = 0;
	const int sAddrTri = 0;
	const int numTri   = 0;
	const int sAddrMat = 0;

#elif SCENE==3		// scene 3 - TRaX cornell box reproduction

	const Color backgroundC( 0.3137f, 0.3529f, 0.3529f );

	MatPrimitive matBoxL( Color( 0.5f, 0.5f, 0.8f ), 0.9f, 0.1f );		// box colors from UCSD
	MatPrimitive matBoxR( Color( 0.8f, 0.5f, 0.5f ), 0.9f, 0.1f );
//	MatPrimitive matBoxL( Color( 0.05f, 0.8f, 0.05f ), 0.7f, 0.3f );		// box colors from UCSD
//	MatPrimitive matBoxR( Color( 0.8f, 0.05f, 0.05f ), 0.7f, 0.3f );
	MatPrimitive matBoxT( Color( 0.8f, 0.8f, 0.8f ), 0.9f, 0.1f );
	MatPrimitive mat1( Color( 1.f, 1.f, 1.f ), 0.f, 0.f, MYRTC::MATETA_PLEXIGLASS, true, true );
	MatPrimitive mat2( Color( 0.8f, 0.8f, 0.8f ), 0.f, 0.f, 1.0f, true, false );
	MatPrimitive mat3( Color( 0.3f, 1.0f, 0.3f ), 0.f, 0.f, 1.f, true, true );

	gpSphere sph1( 0.35f, -0.475f, 0.2f, 0.15f );
	gpSphere sph2( 0.f, -1.0974f, 1.1184f, 0.35f );
	gpSphere sph3( 0.95f, -0.15f, 0.2f, 0.15f );
	//gpSphere sph4( 0.35f, -0.475f, 0.2f, 0.14f );
	gpSphere sph4( 0.95f, -0.15f, 0.2f, 0.13f );
	sph1.m_pMat = &mat1;
	sph2.m_pMat = &mat2;
	sph3.m_pMat = &mat1;
	sph4.m_pMat = &mat3;

	// construct left wall
	gpTriangle l1( Vector(0.0f,-1.0976f,1.1184f), Vector(0.0f,-1.0976f,0.0f), Vector(0.f,0.0f,1.1184f) );
	gpTriangle l2( Vector(0.0f,-1.0976f,0.0f), Vector(0.f,0.f,0.f), Vector(0.f,0.0f,1.1184f) );
	l1.m_pMat = &matBoxL;
	l2.m_pMat = &matBoxL;

	// right wall
	gpTriangle r1( Vector(1.1056f,0.0f,0.f), Vector(1.112f,-1.0976f,0.0f), Vector(1.112f,-1.0976f,1.1184f) );
	gpTriangle r2( Vector(1.112f,-1.0976f,1.1184f), Vector(1.0992f,0.0f,1.1184f), Vector(1.1056f,0.0f,0.f) );
	r1.m_pMat = &matBoxR;
	r2.m_pMat = &matBoxR;

	// top wall
	gpTriangle t1( Vector(0.0f,-1.0976f,0.f), Vector(0.f,-1.0976f,1.1184f), Vector(1.112f,-1.0976f,1.1184f) );
	gpTriangle t2( Vector(1.112f,-1.0976f,0.f), Vector(0.0f,-1.0976f,0.f), Vector(1.112f,-1.0976f,1.1184f) );
	t1.m_pMat = &matBoxT;
	t2.m_pMat = &matBoxT;

	// back wall
	gpTriangle bk1( Vector(1.0992f,0.0f,1.1184f), Vector(1.112f,-1.0976f,1.1184f), Vector(0.f,-1.0976f,1.1184f) );
	gpTriangle bk2( Vector(0.f,-1.0976f,1.1184f), Vector(0.f,0.0f,1.1184f), Vector(1.0992f,0.0f,1.1184f) );
	bk1.m_pMat = &matBoxT;
	bk2.m_pMat = &matBoxT;

	// bottom wall
	MatPrimitive matBoxT2( Color( 0.8f, 0.8f, 0.8f ), 0.7f, 0.3f );
	matBoxT2.m_bFlagProcTileSq = true;
	gpTriangle bt1( Vector(0.f,0.f,0.f), Vector(1.1056f,0.f,0.f), Vector(1.0992f,0.f,1.1184f) );
	gpTriangle bt2( Vector(0.f,0.f,1.1184f), Vector(0.f,0.f,0.f), Vector(1.0992f,0.f,1.1184f) );
	bt1.m_pMat = &matBoxT2;
	bt2.m_pMat = &matBoxT2;

	// tall box
	gpTriangle b1_11( Vector(0.53f,-0.66f,0.592f), Vector(0.846f,-0.66f,0.494f), Vector(0.944f,-0.66f,0.812f) );
	gpTriangle b1_12( Vector(0.628f,-0.66f,0.912f), Vector(0.53f,-0.66f,0.592f), Vector(0.944f,-0.66f,0.812f) );
	gpTriangle b1_21( Vector(0.53f,0.0f,0.592f), Vector(0.846f,0.0f,0.494f), Vector(0.846f,-0.66f,0.494f) );
	gpTriangle b1_22( Vector(0.846f,-0.66f,0.494f), Vector( 0.53f,-0.66f,0.592f), Vector(0.53f,0.0f,0.592f) );
	gpTriangle b1_31( Vector(0.846f,0.0f,0.494f), Vector(0.944f,0.0f,0.812f), Vector(0.944f,-0.66f,0.812f) );
	gpTriangle b1_32( Vector(0.944f,-0.66f,0.812f), Vector(0.846f,-0.66f,0.494f), Vector(0.846f,0.0f,0.494f) );
	gpTriangle b1_41( Vector(0.944f,0.0f,0.812f), Vector(0.628f,0.0f,0.912f), Vector(0.628f,-0.66f,0.912f) );
	gpTriangle b1_42( Vector(0.628f,-0.66f,0.912f), Vector(0.944f,-0.66f,0.812f), Vector(0.944f,0.0f,0.812f) );
	gpTriangle b1_51( Vector(0.628f,-0.66f,0.912f), Vector(0.628f,0.0f,0.912f), Vector(0.53f,0.0f,0.592f) );
	gpTriangle b1_52( Vector(0.53f,0.0f,0.592f), Vector(0.53f,-0.66f,0.592f), Vector(0.628f,-0.66f,0.912f) );
	int tmp2 = 2;
	switch( tmp2 ) {
		case 1:
		default:
		b1_11.m_pMat = &matBoxT;	b1_12.m_pMat = &matBoxT;	b1_21.m_pMat = &matBoxT;
		b1_22.m_pMat = &matBoxT;	b1_31.m_pMat = &matBoxT;	b1_32.m_pMat = &matBoxT;
		b1_41.m_pMat = &matBoxT;	b1_42.m_pMat = &matBoxT;	b1_51.m_pMat = &matBoxT;
		b1_52.m_pMat = &matBoxT;
		break;

	// reflective
		case 2:
		b1_11.m_pMat = &mat2;	b1_12.m_pMat = &mat2;	b1_21.m_pMat = &mat2;	b1_22.m_pMat = &mat2;
		b1_31.m_pMat = &mat2;	b1_32.m_pMat = &mat2;	b1_41.m_pMat = &mat2;	b1_42.m_pMat = &mat2;
		b1_51.m_pMat = &mat2;	b1_52.m_pMat = &mat2;
		break;

	// refractive
		case 3:
		b1_11.m_pMat = &mat1;	b1_12.m_pMat = &mat1;	b1_21.m_pMat = &mat1;	b1_22.m_pMat = &mat1;
		b1_31.m_pMat = &mat1;	b1_32.m_pMat = &mat1;	b1_41.m_pMat = &mat1;	b1_42.m_pMat = &mat1;
		b1_51.m_pMat = &mat1;	b1_52.m_pMat = &mat1;
	}


	// short box
	gpTriangle b2_11( Vector(0.26f,-0.33f,0.13f), Vector(0.58f,-0.33f,0.228f), Vector(0.48f,-0.33f,0.544f) );
	gpTriangle b2_12( Vector(0.164f,-0.33f,0.45f), Vector(0.26f,-0.33f,0.13f), Vector(0.48f,-0.33f,0.544f) );
	gpTriangle b2_21( Vector(0.58f,-0.33f,0.228f), Vector(0.26f,-0.33f,0.13f), Vector(0.26f,0.0f,0.13f) );
	gpTriangle b2_22( Vector(0.26f,0.0f,0.13f), Vector(0.58f,0.0f,0.228f), Vector(0.58f,-0.33f,0.228f) );
	gpTriangle b2_31( Vector(0.58f,0.0f,0.228f), Vector(0.48f,0.0f,0.544f), Vector(0.48f,-0.33f,0.544f) );
	gpTriangle b2_32( Vector(0.48f,-0.33f,0.544f), Vector(0.58f,-0.33f,0.228f), Vector(0.58f,0.0f,0.228f) );
	gpTriangle b2_41( Vector(0.164f,-0.33f,0.45f), Vector(0.48f,-0.33f,0.544f), Vector(0.48f,0.0f,0.544f) );
	gpTriangle b2_42( Vector(0.48f,0.0f,0.544f), Vector(0.164f,0.0f,0.45f), Vector(0.164f,-0.33f,0.45f) );
	gpTriangle b2_51( Vector(0.26f,0.0f,0.13f), Vector(0.26f,-0.33f,0.13f), Vector(0.164f,0.0f,0.45f) );
	gpTriangle b2_52( Vector(0.26f,-0.33f,0.13f), Vector(0.164f,-0.33f,0.45f), Vector(0.164f,0.0f,0.45f) );
	int tmp1 = 3;
	switch( tmp1 ) {
		case 1:
		default:
		b2_11.m_pMat = &matBoxT;	b2_12.m_pMat = &matBoxT;	b2_21.m_pMat = &matBoxT;
		b2_22.m_pMat = &matBoxT;	b2_31.m_pMat = &matBoxT;	b2_32.m_pMat = &matBoxT;
		b2_41.m_pMat = &matBoxT;	b2_42.m_pMat = &matBoxT;	b2_51.m_pMat = &matBoxT;
		b2_52.m_pMat = &matBoxT;
		break;

	// reflective
		case 2:
		b2_11.m_pMat = &mat2;	b2_12.m_pMat = &mat2;	b2_21.m_pMat = &mat2;	b2_22.m_pMat = &mat2;
		b2_31.m_pMat = &mat2;	b2_32.m_pMat = &mat2;	b2_41.m_pMat = &mat2;	b2_42.m_pMat = &mat2;
		b2_51.m_pMat = &mat2;	b2_52.m_pMat = &mat2;
		break;

	// refractive
		case 3:
		b2_11.m_pMat = &mat1;	b2_12.m_pMat = &mat1;	b2_21.m_pMat = &mat1;	b2_22.m_pMat = &mat1;
		b2_31.m_pMat = &mat1;	b2_32.m_pMat = &mat1;	b2_41.m_pMat = &mat1;	b2_42.m_pMat = &mat1;
		b2_51.m_pMat = &mat1;	b2_52.m_pMat = &mat1;
		break;
	}

	MatPrimitive matLt( Color( 20.f, 20.f, 20.f ), 0.f, 1.f );
	gpTriangle lt1( Vector(0.686f,-1.0974f,0.454f), Vector(0.426f,-1.0974f,0.454f), Vector(0.426f,-1.0974f,0.664f) );
	gpTriangle lt2( Vector(0.686f,-1.0974f,0.664f), Vector(0.686f,-1.0974f,0.454f), Vector(0.426f,-1.0974f,0.664f) );
	lt1.m_pMat = &matLt;
	lt2.m_pMat = &matLt;

	const int numSph   = 3;
	gpSphere* objSph[] = { &sph1, &sph2, &sph3, &sph4 };

	const int numTriO    = 32;
	gpTriangle* objTri[] = { &l1, &l2, &r1, &r2, &t1, &t2, &bk1, &bk2, &bt1, &bt2, &lt1, &lt2,
							&b1_11, &b1_12, &b1_21, &b1_22, &b1_31, &b1_32, &b1_41, &b1_42, &b1_51, &b1_52, 
							&b2_11, &b2_12, &b2_21, &b2_22, &b2_31, &b2_32, &b2_41, &b2_42, &b2_51, &b2_52 };

	myCamera cam( Vector(0.5528f, -0.549f, -1.57f), Vector(0.5528f,-0.549f,0.f), Vector(0.f,-1.f,0.f),
					width, height, MYRTC::CAM_FOV_193_DEG );

	const int numPtLights = 1;
	Color ltAmb( 0.6f, 0.6f, 0.6f );
	ltPoint ltPt1( Vector(0.5528f, -1.095f, 0.5592f), Color(1.f, 1.f, 1.f) );
	ltPoint* ptLts[] = { &ltPt1 };

	const int maxRecursionDepth = 15;

	const float kDiff  = 0.7f;
	const float kAmb   = 0.3f;
	const int sAddrBVH = 0;
	const int sAddrTri = 0;
	const int numTri   = 0;
	const int sAddrMat = 0;
#endif

	// set up scene construct
	RayPrimitive ray( cam.getEye() );
	const float iPrem = 2.f * widthInv;
	const float iSub  = 1.f - widthInv;			// to make things match -> need to switch - to +
	const float jPrem = 2.f * heightInv;
	const float jSub  = 1.f - heightInv;

	const unsigned int nSampleInDim = 2;		// how many samples in x or y dimensions to take (jittered strategy)
	const float numSampleInv		= 1.f / nSampleInDim;
	const float numSampleInvSq		= numSampleInv*numSampleInv;
	const float maxRandInv			= 1.f/RAND_MAX;

	// Iterate over all of the pixels
	// Using atomicinc here allows your code to have only one thread execute per return value
	for( int pixel = atomicinc(0); pixel < width * height; pixel = atomicinc(0) ) {

		// Get screen location
		int i = pixel % heightINT;
		int j = pixel / widthINT;

		// do we jitter samples?
	#if SUPERSAMPLE==1
		Color pxlColor(0.f,0.f,0.f);
		for( unsigned int sx=0; sx<nSampleInDim; ++sx ) {
			for( unsigned int sy=0; sy<nSampleInDim; ++sy ) {

		// lower-left location of sample
		const float iLower = i*iPrem - iSub;
		const float jLower = j*jPrem - jSub;

		// get x,y locations for samples
		const float sample1 = ( sx + rand()*maxRandInv ) * numSampleInv;
		const float sample2 = ( sy + rand()*maxRandInv ) * numSampleInv;

		// 1. My camera
		cam.generateRayDirection( ray, iLower + sample1*widthInv, jLower + sample2*heightInv );
	#else

		// 1. My camera
		cam.generateRayDirection( ray, i*iPrem - iSub, j*jPrem - jSub );
	#endif


		// 2. intersect with intersection records
		IntersectionRecord closestIr;
		ray.m_fMinT = MYRTC::RAY_EPS;
		ray.m_fMaxT = MYRTC::RAY_INF;

	#if DO_BVH==0
		gpTriangle tmpT;
		int tID			 = 0;
		bool intersected = false;
		for( int tAddr=sAddrTri; tID<numTri; ++tID, tAddr+=gpTriangle::SIZE_IN_TRAX ) {
			tmpT.set( tAddr );
			intersected |= tmpT.intersectDetailed( ray, closestIr );
		}

		// 3. shade point but only if on the correct side
		if( intersected ) {
	#else
	#if SCENE==1
		bool intersected = false;
		for( int s=0; s<numSph; ++s )
			intersected |= objSph[s]->intersectDetailed( ray, closestIr );
		intersected |= intersectDetailedWithBVH( ray, closestIr, sAddrBVH );
		
		// 3. now shade
		if( intersected ) {
	#elif SCENE==2 || SCENE==3
		bool intersected = false;
		for( int s=0; s<numSph; ++s )
			intersected |= objSph[s]->intersectDetailed( ray, closestIr );
		for( int t=0; t<numTriO; ++t )
			intersected |= objTri[t]->intersectDetailed( ray, closestIr );

		// 3. now shade
		if( intersected ) {
	#else
		// 3. shade point
/*		printf( "bvh traversal\n");
		bool intersected =  intersectDetailedWithBVH( ray, closestIr, sAddrBVH );
		printf( "survived\n");
		if( intersected ) {
*/
		if( intersectDetailedWithBVH( ray, closestIr, sAddrBVH ) ) {
	#endif
	#endif

			// ambient light computation
			Color outClr( ltAmb );
		#if SCENE==TRAX_SCENE || SCENE==4			// TRaX
			//MatPrimitive tmpM( sAddrMat, closestIr.m_iGeomMatID, kDiff, kAmb );
			//outClr.mult( tmpM.m_fKamb );
			outClr.mult( kAmb );		// for now, we can save some cycles
		#elif SCENE==1
			if( closestIr.m_pGeomMat==0 )
				outClr.mult( kAmb );		// for now, we can save some cycles
			else
				outClr.mult( closestIr.m_pGeomMat->m_fKamb );
		#else
			outClr.mult( closestIr.m_pGeomMat->m_fKamb );
		#endif

			// get intersection point/normal
			Vector ptLoc, ptNorm;
			ray.getPoint( ptLoc, ray.m_fMaxT );

		#if SCENE==1 || SCENE==2 || SCENE==3
			if( closestIr.m_iGeomHitAddr>0 ) {
				gpTriangle tmpT;
				tmpT.set( closestIr.m_iGeomHitAddr );
				closestIr.m_iGeomMatID = tmpT.m_matID;
				tmpT.getNormal( ptNorm, ptLoc );
			}
			else if( closestIr.m_oType==MYRTC::OT_SPHERE ) {
				static_cast<const gpSphere*>(closestIr.m_pGeomHit)->getNormal( ptNorm, ptLoc );
				closestIr.m_iGeomMatID = -1;
			}
			else {
				static_cast<const gpTriangle*>(closestIr.m_pGeomHit)->getNormal( ptNorm, ptLoc );
				closestIr.m_iGeomMatID = -1;
			}
		#else		// TRAX
			gpTriangle tmpT;
			tmpT.set( closestIr.m_iGeomHitAddr );
			closestIr.m_iGeomMatID = tmpT.m_matID;
			tmpT.getNormal( ptNorm, ptLoc );
		#endif

			RayPrimitive shRay( ptLoc );
			shRay.m_fMinT = MYRTC::RAY_INTER_STEP;

			// look through all lights in scene and accumulate shading
		#if SCENE==TRAX_SCENE	// TRaX
			{	// get shadow ray to light
				Vector intToLight( ptLt.m_vLocation );
		#else
			for( unsigned int l=0; l<numPtLights; ++l ) {
				// get shadow ray to light
				Vector intToLight( ptLts[l]->m_vLocation );
		#endif
				intToLight.sub( shRay.m_vOrigin );
				shRay.m_vDirection.set( intToLight );
				shRay.m_fMaxT = 1.0f;

				// intersect shadow ray with scene -> slower
			#if DO_BVH==0
				bool outShadow  = true;
				int tID			= 0;
				for( int tAddr=sAddrTri; tID<numTri && outShadow; ++tID, tAddr+=gpTriangle::SIZE_IN_TRAX ) {
					tmpT.set( tAddr );
					outShadow = tmpT.intersectQuick( shRay );
				}

				// do we need to shade?
				if( outShadow ) {
			#else
			#if SCENE==1
				bool outShadow = true;
				for( int s=0; s<numSph && outShadow; ++s )
					outShadow = objSph[s]->intersectQuick( shRay );

				if( outShadow )
					outShadow = intersectQuickWithBVH( shRay, sAddrBVH );

				if( outShadow ) {
			#elif SCENE==2 || SCENE==3
				bool outShadow = true;
				for( int s=0; s<numSph && outShadow; ++s )
					outShadow = objSph[s]->intersectQuick( shRay );
				for( int t=0; t<numTriO && outShadow; ++t )
					outShadow = objTri[t]->intersectQuick( shRay );

				if( outShadow ) {
			#else
/*				printf( "shadows\n" );
				bool tmp = intersectQuickWithBVH( shRay, sAddrBVH );
				printf( "woot!\n");
				if(tmp ) {
*/
				if( intersectQuickWithBVH( shRay, sAddrBVH ) ) {
			#endif
			#endif

					// get cos-varied amount of light reaching the point (for some reason backfacing was an issue :( )
				#if SCENE==1 || SCENE==3 || SCENE==4
					const float cosPhi = max( ptNorm.dot( shRay.m_vDirection ) * invsqrt( shRay.m_vDirection.lengthSq() ), 0.f );
				#else
					const float normDot = ptNorm.dot( shRay.m_vDirection );
					const float cosPhi = (normDot > 0.f ? normDot : -normDot ) * invsqrt( shRay.m_vDirection.lengthSq() );
				#endif

					// light we get can be seen -> do lambertian (diffuse shading)
				#if SCENE==TRAX_SCENE		// TRAX
					Color tmpClr( ptLt.m_clrAmbient );
					//tmpClr.mult( (cosPhi*tmpM.getKDiff() ) );
					// for now can save some cycles
					tmpClr.mult( (cosPhi*kDiff) );
				#elif SCENE==1
					Color tmpClr( ptLt.m_clrAmbient );
					if( closestIr.m_iGeomMatID>=0 )
						tmpClr.mult( (cosPhi*kDiff) );
					else
						tmpClr.mult( (cosPhi*closestIr.m_pGeomMat->m_fKdiff) );
				#elif SCENE==4
					Color tmpClr( ptLts[l]->m_clrAmbient );
					//tmpClr.mult( (cosPhi*tmpM.getKDiff() ) );
					tmpClr.mult( (cosPhi*kDiff) );
				#else
					Color tmpClr( ptLts[l]->m_clrAmbient );
					tmpClr.mult( (cosPhi*closestIr.m_pGeomMat->m_fKdiff) );
				#endif
					outClr.add( tmpClr );
				}

			} // end l

			// compute diffuse material color
		#if SCENE==TRAX_SCENE
			MatPrimitive tmpM( sAddrMat, closestIr.m_iGeomMatID );
			outClr.mult( tmpM.m_clrDiff );
		#elif SCENE==1
			Color matDiffClr;
			if( closestIr.m_iGeomMatID>=0 ) {
				// do we tile?
				if( closestIr.m_iGeomMatID!=matIDToTile ) {
					MatPrimitive tmpM( sAddrMat, closestIr.m_iGeomMatID );
					//outClr.mult( tmpM.m_clrDiff );
					matDiffClr.set( tmpM.m_clrDiff );
				}
				else {
					int i = static_cast<int>( ptLoc.x*tileScale );
					int j = static_cast<int>( ptLoc.y*tileScale );
					int k = static_cast<int>( ptLoc.z*tileScale );
					int cell = ( i + j + k )%2;
					if( cell==0 )
						matDiffClr.set( evenClr );
					else
						matDiffClr.set( oddClr );
				}
			}
			else if( closestIr.m_pGeomMat ){
				//outClr.mult( closestIr.m_pGeomMat->m_clrDiff );
				matDiffClr.set( closestIr.m_pGeomMat->m_clrDiff );
			}
		#elif SCENE==2
			Color matDiffClr( closestIr.m_pGeomMat->m_clrDiff );
		#elif SCENE==3
			Color matDiffClr( closestIr.m_pGeomMat->m_clrDiff );
			// do procedural squares
			if( closestIr.m_pGeomMat->m_bFlagProcTileSq ) {
				int i = static_cast<int>( ptLoc.x*tileScale );
				int j = static_cast<int>( ptLoc.y*tileScale );
				int k = static_cast<int>( ptLoc.z*tileScale );
				int cell = ( i + j + k )%2;
				if( cell==0 )
					matDiffClr.set( evenClr );
				else
					matDiffClr.set( oddClr );
			}
		#elif SCENE==4
			Color matDiffClr;
			// do we tile?
			if( closestIr.m_iGeomMatID!=matIDToTile ) {
				MatPrimitive tmpM( sAddrMat, closestIr.m_iGeomMatID );
				//outClr.mult( tmpM.m_clrDiff );
				matDiffClr.set( tmpM.m_clrDiff );
			}
			else {
				int i = static_cast<int>( ptLoc.x*tileScale );
				int j = static_cast<int>( ptLoc.y*tileScale );
				int k = static_cast<int>( ptLoc.z*tileScale );
				int cell = ( i + j + k )%2;
				if( cell==0 )
					matDiffClr.set( evenClr );
				else
					matDiffClr.set( oddClr );
			}
		#else
			outClr.mult( closestIr.m_pGeomMat->m_clrDiff );
		#endif

//			printf("prereflections\n");


		#if (DO_REFLECTIONS==1) &&( SCENE==1 || SCENE==2 || SCENE==3 || SCENE==4)
			// set-up REFLECTION function call (test we're reflecting on the outside?)
			if( closestIr.m_iGeomMatID==matIDToReflect || closestIr.m_iGeomMatID==matIDToReflect2 ||
				closestIr.m_iGeomMatID==matIDToRefract ||
				(closestIr.m_pGeomMat && closestIr.m_pGeomMat->m_bFlagReflective) ) {

				// compute refracted direction
				Vector refDir;
				const float viewDotN = ray.m_vDirection.dot(ptNorm);
				const float thisEta  = 1.0f;
				const float thatEta  = (closestIr.m_pGeomMat) ? closestIr.m_pGeomMat->m_fMatEta : refrEta;
				float etaFrac		 = thisEta / thatEta;
				float goOutsideIn    = 1.f;
				if( viewDotN > 0.f ) {
					etaFrac		= thatEta / thisEta;
					goOutsideIn = -1.f;
				}
				const bool TIR = computeTransmittedDirection( refDir, ray.m_vDirection, ptNorm, viewDotN, etaFrac );//thisEta, thatEta );

				if( TIR ) {
					outClr.set( 1.f );
					continue;
				}

				// compute reflected direction
				shRay.m_vDirection.set( ptNorm );
				shRay.m_vDirection.mult( (-2.f)*viewDotN );
				shRay.m_vDirection.add( ray.m_vDirection );
				shRay.m_vDirection.normalize();

				// move the intersection (start ray point) away from surface  // todo -> remove!
				shRay.m_vOrigin.set( ptNorm );
				shRay.m_vOrigin.mult( goOutsideIn*RAY_OFFSET );
				shRay.m_vOrigin.add( ptLoc );
				RayPrimitive shRay2;

				// compute reflected coefficient
				float c = viewDotN<0.f ? -viewDotN : viewDotN;//refDir.dot(ptNorm) ;

				// recurse through the scene
				Color refl( 0.f, 0.f, 0.f );
				const bool thisObjRefractive = (closestIr.m_iGeomMatID==matIDToRefract) || (closestIr.m_pGeomMat && closestIr.m_pGeomMat->m_bFlagRefractive);
				ComputeRecursiveRays( refl, shRay, shRay2, closestIr, 1, numSph, objSph, 
											numTriO, objTri, ltAmb, numPtLights, ptLts, 
											sAddrBVH, sAddrMat, 
											maxRecursionDepth, kAmb, kDiff );

				// do we need to scale this reflected bit?
				if( thisObjRefractive==false ) {
					outClr.set( refl );
				}
				else {

					float r0, r = 1.f - c;

					// scale output color based on Schlick approximation for Fresnel reflectance	
					r0 = ( thisEta-thatEta )/( thisEta+thatEta );
					r0 *= r0;
					const float r2 = r*r;
					r *= r2*r2;
					r  = r0 + (1.f-r0) * r;
					refl.mult( r );
					outClr.set( refl );

					// did we hit total internal reflection?
					if( TIR==false ) {

						// compute refraction
						shRay.m_vDirection.set( refDir );
						shRay.m_vDirection.normalize();

						// move the intersection (start ray point) away from surface -> todo: remove
						shRay.m_vOrigin.set( ptNorm );
						shRay.m_vOrigin.mult( goOutsideIn*(-RAY_OFFSET) );
						shRay.m_vOrigin.add( ptLoc );

						// do the computation
						refl.set( 0.f, 0.f, 0.f );
						ComputeRecursiveRays( refl, shRay, shRay2, closestIr, 1, numSph, objSph, 
												numTriO, objTri, ltAmb, numPtLights, ptLts, sAddrBVH, sAddrMat,
												maxRecursionDepth, kAmb, kDiff );

						refl.mult( 1.f-r );
						outClr.add( refl );
					}
//					else
//						outClr.set( 1.f,1.f,0.f);
				}
			}
		#endif

		#if SCENE==1 || SCENE==2 || SCENE==3 || SCENE==4
			outClr.mult( matDiffClr );
		#endif

	#if SUPERSAMPLE==1
			pxlColor.add( outClr );
		} // end intersected
		else {
			pxlColor.set( backgroundC );
		}

		}// end sy
		}// end sx

		pxlColor.mult( numSampleInvSq );
		pxlColor.clamp01();
		StorePixelIntoFB( framebuffer, pixel, pxlColor );

	#else
			outClr.clamp01();
			StorePixelIntoFB( framebuffer, pixel, outClr );
			continue;
		} // end intersected

		// background color
		StorePixelIntoFB( framebuffer, pixel, backgroundC );
	#endif
	}

	// Conditional to only execute the following on the CPU version
#if TRAX==0
	printf( "Done drawing.\n" );
#endif

	// Cleanup code. On the CPP target writes out.png.
//	trax_cleanup();
	return 0;
}




/// Helper function that helps us to store a pixel into the frame buffer
void StorePixelIntoFB( const int &fb, const int &pixel, const Color &clr ) {
	int base = fb + pixel*3;
	storef( clr.r, base, 0 );
	storef( clr.g, base, 1 );
	storef( clr.b, base, 2 );
}


/**
 * Helper function that throws a ray through our scene and traverses
 *   the BVH, while keeping track of the intersection record
 * NOTE: this works only for the TRaX triangle-only scenes
 *
 * @param ray    - ray to cast through the BVH (will be updated with far t-value)
 * @param intRec - record holding intersection information
 */
inline bool intersectDetailedWithBVH( RayPrimitive &ray, IntersectionRecord &intRec, const int &sAddrBVH ) {

	int bvhStack[BVH_TRAVERSAL_STACK_SIZE];
	ray.computeBoxIntersectInfo();
	
	bool intersected = false;
	int stackOffset  = 0, nodeId = 0;
	float t0, t1;
	
	gpBox      tmpB;
	gpTriangle tmpT;

	while( true ) {

		// load BVH box
		int tmpAddr = sAddrBVH + gpBox::SIZE_IN_TRAX*nodeId;
		tmpB.set( tmpAddr );

		// intersect with ray
		if( tmpB.intersectDetailed( t0, t1, ray ) ) {

			// are we a parent node?
			int numChildren = loadi( tmpAddr, 6 );
			if( numChildren < 0 ) {
			/*	nodeId					= loadi( tmpAddr, 7 );	// traverse left child
				bvhStack[stackOffset++] = nodeId+1;				// push right child
				continue;
			//*/
				// check which child to add first based on which is closer
				// left is closer
				if( ( ray.m_iDirSignX && tmpB.pts[1].x>loadf(tmpAddr,0) ) || 
					( ray.m_iDirSignY && tmpB.pts[1].y>loadf(tmpAddr,1) ) ||
					( ray.m_iDirSignZ && tmpB.pts[1].z>loadf(tmpAddr,2) ) ) {
					nodeId					= loadi( tmpAddr, 7 );
					bvhStack[stackOffset++] = nodeId+1;
					continue;
				}
				// right is closer
				else {
					nodeId					= loadi( tmpAddr, 7 );
					bvhStack[stackOffset++] = nodeId++;
					continue;
				}
			}

			// are we leaf?
			else {

				//intRec.m_iGeomHitAddr = nodeId;
				//return true;

				// intersect with all interior children
				int tAddr = loadi( tmpAddr, 7 );
				int tID	  = 0;
				for( ; tID<numChildren; ++tID, tAddr+=gpTriangle::SIZE_IN_TRAX ) {
					tmpT.setQuick( tAddr );
					intersected |= tmpT.intersectDetailed( ray, intRec );
				}

				// pop current box off stack
//				if( stackOffset==0 )
//					break;
//				nodeId = bvhStack[--stackOffset];
			}
		}
//		else {

			// no intersection -> pop current box off stack
			if( stackOffset==0 )
				break;
			nodeId = bvhStack[--stackOffset];
//		}
	}
	/**/
	return intersected;
}


/**
 * Quick intersection of a ray with the BVH scene. This is intended for 
 *   shadow rays only. Returns whether a ray has hit anything
 * NOTE: this works only for the TRaX triangle-only scenes
 *
 * @param ray - ray to cast through the BVH (will be updated with far t-value)
 * @return flag whether ray hits anything
 */
inline bool intersectQuickWithBVH( RayPrimitive &ray, const int &sAddrBVH ) {

	int bvhStack[BVH_TRAVERSAL_STACK_SIZE];
	ray.computeBoxIntersectInfo();
	
	int stackOffset = 0, nodeId = 0;
	float t0, t1;
	
	gpBox      tmpB;
	gpTriangle tmpT;

	while( true ) {

		// load BVH box
		int tmpAddr = sAddrBVH + gpBox::SIZE_IN_TRAX*nodeId;
		tmpB.set( tmpAddr );

		// intersect with ray
		if( tmpB.intersectDetailed( t0, t1, ray ) ) {

			// are we a parent node?
			int numChildren = loadi( tmpAddr, 6 );
			if( numChildren < 0 ) {
				// check which child to add first based on which is closer
				// left is closer
				if( ( ray.m_iDirSignX && tmpB.pts[1].x>loadf(tmpAddr,0) ) || 
					( ray.m_iDirSignY && tmpB.pts[1].y>loadf(tmpAddr,1) ) ||
					( ray.m_iDirSignZ && tmpB.pts[1].z>loadf(tmpAddr,2) ) ) {
					nodeId					= loadi( tmpAddr, 7 );
					bvhStack[stackOffset++] = nodeId+1;
					continue;
				}
				// right is closer
				else {
					nodeId					= loadi( tmpAddr, 7 );
					bvhStack[stackOffset++] = nodeId++;
					continue;
				}
			}

			// are we leaf?
			else {

				// intersect with all interior children
				int tAddr = loadi( tmpAddr, 7 );
				int tID	  = 0;
				for( ; tID<numChildren; ++tID, tAddr+=gpTriangle::SIZE_IN_TRAX ) {
					tmpT.setQuick( tAddr );
					if( tmpT.intersectQuick( ray )==false )
						return false;
				}
			}
		}
		// no intersection -> pop current box off stack
		if( stackOffset==0 )
			break;
		nodeId = bvhStack[--stackOffset];
	}
	return true;
}


// return true if TIR
inline bool computeTransmittedDirection( Vector &tDir, const Vector &inDir, const Vector &norm, 
										const float &dDotN, const float &etaFrac ) {


	// redoing test
	const float nDnt  = etaFrac;
	const float inner = nDnt*nDnt*( 1.f - dDotN*dDotN );
	if( inner > 1.f ) {
		return true;
	}
	tDir.set( norm );
	if( dDotN < 0.f )
		tDir.mult( -nDnt*dDotN - sqrt(1.f-inner) );
	else
		tDir.mult( -nDnt*dDotN + sqrt(1.f-inner) );

	Vector tmpV( inDir );
	tmpV.mult( nDnt );
	tDir.add( tmpV );
	return false;
}

inline void ComputeRecursiveRays( Color &cOut, RayPrimitive &ruPRay, RayPrimitive &ruSRay, IntersectionRecord &closestIr,
								 const unsigned int &level, const int &numSph, gpSphere ** spheres,
								 const int &numTri, gpTriangle ** tris, const Color &ambLt,
								 const int &numLt, ltPoint ** ptLts, const int &sAddrBVH, const int &sAddrMat,
								 const unsigned int &maxLvl, const float &kAmb, const float &kDiff ) {

	// 0. check depth
	if( level>maxLvl )
		return;

	// 1. intersect with scene
	bool intersected = false;
	ruPRay.m_fMinT = RAY_OFFSET;
	ruPRay.m_fMaxT = MYRTC::RAY_INF;

	for( int p=0; p<numSph; ++p ) {
		intersected |= spheres[p]->intersectDetailed( ruPRay, closestIr );
	}
	for( int p=0; p<numTri; ++p ) {
		intersected |= tris[p]->intersectDetailed( ruPRay, closestIr );
	}
#if SCENE==TRAX_SCENE || SCENE==1 || SCENE==4
	intersected |= intersectDetailedWithBVH( ruPRay, closestIr, sAddrBVH );
#endif

	// 2. shade point
	if( intersected ) {

		// get intersection point/normal
		Vector ptLoc, ptNorm;
		ruPRay.getPoint( ptLoc, ruPRay.m_fMaxT );

		if( closestIr.m_iGeomHitAddr>0 ) {
			gpTriangle tmpT;
			tmpT.set( closestIr.m_iGeomHitAddr );
			closestIr.m_iGeomMatID = tmpT.m_matID;
			tmpT.getNormal( ptNorm, ptLoc );
		}
		else {
			switch( closestIr.m_oType ) {

				// sphere!
				case MYRTC::OT_SPHERE:
					static_cast<const gpSphere*>(closestIr.m_pGeomHit)->getNormal( ptNorm, ptLoc );
					break;

				// triangle
				case MYRTC::OT_TRIANGLE:
					static_cast<const gpTriangle*>(closestIr.m_pGeomHit)->getNormal( ptNorm, ptLoc );
					break;

				// some random object -> can't intersect :/
				default:
					break;
			};
		}


		// are we not reflective? -> shade, otherwise reflect
		if( (closestIr.m_pGeomMat==0 || closestIr.m_pGeomMat->m_bFlagReflective==false) &&
			closestIr.m_iGeomMatID!=matIDToReflect && closestIr.m_iGeomMatID!=matIDToReflect2
			&& closestIr.m_iGeomMatID!=matIDToRefract ) {

			// ambient light computation
			cOut.set( ambLt );
			if( closestIr.m_pGeomMat )
				cOut.mult( closestIr.m_pGeomMat->m_fKamb );
			else
				cOut.mult( kAmb );

			// loop over lights and see if we have any color
			ruSRay.m_vOrigin.set( ptLoc );
			ruSRay.m_fMinT = MYRTC::RAY_INTER_STEP;

			for( int l=0; l<numLt; ++l ) {

				// get shadow ray to light
				Vector intToLight( ptLts[l]->m_vLocation );
				intToLight.sub( ruSRay.m_vOrigin );
				ruSRay.m_vDirection.set( intToLight );
				ruSRay.m_fMaxT = 1.0f;

				// intersect shadow ray with scene -> slower
				bool outShadow = true;
				for( int p=0; p<numSph && outShadow; ++p ) {
					outShadow = spheres[p]->intersectQuick( ruSRay );
				}
				for( int p=0; p<numTri && outShadow; ++p ) {
					outShadow = tris[p]->intersectQuick( ruSRay );
				}
			#if SCENE==TRAX_SCENE || SCENE==1 || SCENE==4
				if( outShadow )
					outShadow = intersectQuickWithBVH( ruSRay, sAddrBVH );
			#endif

				// do we need to shade?
				if( outShadow ) {

					// get cos-varied amount of light reaching the point
					if( closestIr.m_iGeomHitAddr==0 ) {
						const float cosPhi = max( ptNorm.dot( ruSRay.m_vDirection ) * invsqrt( ruSRay.m_vDirection.lengthSq() ), 0.f );
						Color tmpClr( ptLts[l]->m_clrAmbient );
						tmpClr.mult( (cosPhi*closestIr.m_pGeomMat->m_fKdiff) );
						cOut.add( tmpClr );
					}
					else {
					#if SCENE==4
						const float cosPhi = max( ptNorm.dot( ruSRay.m_vDirection ) * invsqrt( ruSRay.m_vDirection.lengthSq() ), 0.f );
					#else
						const float normDot = ptNorm.dot( ruSRay.m_vDirection );
						const float cosPhi = (normDot > 0.f ? normDot : -normDot ) * invsqrt( ruSRay.m_vDirection.lengthSq() );
					#endif
						Color tmpClr( ptLts[l]->m_clrAmbient );
						tmpClr.mult( (cosPhi*kDiff) );
						cOut.add( tmpClr );
					}
				}

			} // end l

			// apply diffuse material color
		#if SCENE==TRAX_SCENE
			MatPrimitive tmpM( sAddrMat, closestIr.m_iGeomMatID );
			cOut.mult( tmpM.m_clrDiff );
		#elif SCENE==1
			if( closestIr.m_iGeomMatID>=0 ) {
				if( closestIr.m_iGeomMatID!=matIDToTile ) {
					MatPrimitive tmpM( sAddrMat, closestIr.m_iGeomMatID );
					cOut.mult( tmpM.m_clrDiff );
				}
				else {
					int i = static_cast<int>( ptLoc.x*tileScale );
					int j = static_cast<int>( ptLoc.y*tileScale );
					int k = static_cast<int>( ptLoc.z*tileScale );
					int cell = ( i + j + k )%2;
					if( cell==0 )
						cOut.mult( evenClr );
					else
						cOut.mult( oddClr );
				}
			}
			else if( closestIr.m_pGeomMat )
				cOut.mult( closestIr.m_pGeomMat->m_clrDiff );
		#elif SCENE==3
			// do procedural squares
			if( closestIr.m_pGeomMat->m_bFlagProcTileSq ) {
				int i = static_cast<int>( ptLoc.x*tileScale );
				int j = static_cast<int>( ptLoc.y*tileScale );
				int k = static_cast<int>( ptLoc.z*tileScale );
				int cell = ( i + j + k )%2;
				if( cell==0 )
					cOut.mult( evenClr );
				else
					cOut.mult( oddClr );
			}
			else
				cOut.mult( closestIr.m_pGeomMat->m_clrDiff );
		#elif SCENE==4
			if( closestIr.m_iGeomMatID!=matIDToTile ) {
				MatPrimitive tmpM( sAddrMat, closestIr.m_iGeomMatID );
				cOut.mult( tmpM.m_clrDiff );
			}
			else {
				int i = static_cast<int>( ptLoc.x*tileScale );
				int j = static_cast<int>( ptLoc.y*tileScale );
				int k = static_cast<int>( ptLoc.z*tileScale );
				int cell = ( i + j + k )%2;
				if( cell==0 )
					cOut.mult( evenClr );
				else
					cOut.mult( oddClr );
			}
		#else
			cOut.mult( closestIr.m_pGeomMat->m_clrDiff );
		#endif
			return;
		}

		// get diffuse color for this reflective/refractive object...
		Color matDiffClr( 1.f, 1.f, 1.f );
		if( closestIr.m_pGeomMat )
			matDiffClr.set( closestIr.m_pGeomMat->m_clrDiff );
		else {
			MatPrimitive tmpM( sAddrMat, closestIr.m_iGeomMatID );
			matDiffClr.set( tmpM.m_clrDiff );
		}


		// set-up REFLECTION function call (test we're reflecting on the outside?)
		// compute refracted direction
		Vector refDir;
		const float viewDotN = ruPRay.m_vDirection.dot(ptNorm);
		float thisEta		 = 1.f;//ruPRay.m_fCurEta;//1.0f;
		float thatEta		 = (closestIr.m_pGeomMat) ? closestIr.m_pGeomMat->m_fMatEta : refrEta;
		float etaFrac		 = thisEta / thatEta;
		float goOutsideIn    = 1.f;
		if( viewDotN > 0.f ) {
			// we're going outside of the material. Need to figure out what is the "thatEta"
			// do this only for scene 3, where we have a nested dielectric material
			etaFrac     = thatEta / thisEta;
			goOutsideIn = -1.f;
		}
		const bool TIR = computeTransmittedDirection( refDir, ruPRay.m_vDirection, ptNorm, viewDotN, etaFrac );

		// compute reflected direction
		ruSRay.m_vDirection.set( ptNorm );
		ruSRay.m_vDirection.mult( (-2.f)*viewDotN );
		ruSRay.m_vDirection.add( ruPRay.m_vDirection );
		ruSRay.m_vDirection.normalize();

		// move the intersection (start ray point) away from surface
		ruSRay.m_vOrigin.set( ptNorm );
		ruSRay.m_vOrigin.mult( goOutsideIn*RAY_OFFSET );
		ruSRay.m_vOrigin.add( ptLoc );

		// compute reflected coefficient
		float c = viewDotN<0.f ? -viewDotN : viewDotN;//refDir.dot(ptNorm) ;

		// recurse through the scene
		Color refl( 0.f, 0.f, 0.f );
		//const bool thisObjRefractive = closestIr.m_pGeomMat->m_bFlagRefractive;
		const bool thisObjRefractive = (closestIr.m_iGeomMatID==matIDToRefract) || (closestIr.m_pGeomMat && closestIr.m_pGeomMat->m_bFlagRefractive);
		ComputeRecursiveRays( refl, ruSRay, ruPRay, closestIr, level+1, numSph, spheres, 
									numTri, tris, ambLt, numLt, ptLts, sAddrBVH, sAddrMat,
									maxLvl, kAmb, kDiff );

		// do we need to scale this reflected bit?
		if( thisObjRefractive==false ) {
			cOut.add( refl );
		}
		else {
			float r0, r = 1.f - c;

			// scale output color based on Schlick approximation for Fresnel reflectance
			r0  = ( thisEta-thatEta )/( thisEta+thatEta );
			r0 *= r0;
			const float r2 = r*r;
			r *= r2*r2;
			r  = r0 + (1.f-r0) * r;

			// check if we're in total internal reflection
			if( TIR ) {
				cOut.set( refl );
				cOut.mult( matDiffClr );
				return;
			}
			refl.mult( r );
			cOut.set( refl );

			// compute refraction
			ruSRay.m_vDirection.set( refDir );
			ruSRay.m_vDirection.normalize();

			// move the intersection (start ray point) away from surface
			ruSRay.m_vOrigin.set( ptNorm );
			ruSRay.m_vOrigin.mult( goOutsideIn*(-RAY_OFFSET) );
			ruSRay.m_vOrigin.add( ptLoc );
			ruSRay.m_fMaxT   = MYRTC::RAY_INF;
			//ruSRay.m_fCurEta = thatEta;

			// do the computation
			refl.set( 0.f, 0.f, 0.f );
			ComputeRecursiveRays( refl, ruSRay, ruPRay, closestIr, level+1, numSph, spheres, 
									numTri, tris, ambLt, numLt, ptLts, sAddrBVH, sAddrMat,
									maxLvl, kAmb, kDiff );

			refl.mult( 1.f-r );
			cOut.add( refl );
			cOut.mult( matDiffClr );
		}
	}
}

