// defining some functions Windows is missing
#include "commonFcns.h"

#include "pthread.h"

#include "LoadMemory.h"
#include "ReadLightfile.h"
#include "ReadViewfile.h"
// #include "LoadMemory.cc"
// #include "OBJLoader.cc"
// #include "IWLoader.cc"
// #include "BVH.cc"
// #include "Grid.cc"
// #include "Camera.cc"
// #include "ReadLightfile.cc"
// #include "ReadViewfile.cc"
// #include "Material.cc"
// #include "Triangle.cc"
// #include "MTLLoader.cc"
// #include "PPM.cc"

#ifndef WIDTH
#define WIDTH 512
#endif
#ifndef HEIGHT
#define HEIGHT 512
#endif

unsigned int image_width = WIDTH;
unsigned int image_height = HEIGHT;

/// Pthreads stuff
const unsigned int numAtomicInc = 8;
pthread_t *threads = 0;
pthread_mutex_t mainMutex[ numAtomicInc ];


// C versions of everything
FourByte * trax_memory_pointer;

int loadi( int base, int offset = 0 ) {
  return trax_memory_pointer[base+offset].ivalue;
}
float loadf( int base, int offset = 0 ) {
  return trax_memory_pointer[base+offset].fvalue;
}
void storei( int value, int base, int offset = 0 ) {
  trax_memory_pointer[base+offset].ivalue = value;
}
void storef( float value, int base, int offset = 0 ) {
  trax_memory_pointer[base+offset].fvalue = value;
}

// Arithmetic
int atomicinc( int location) {
/*	// original:
  static unsigned int val[] = {0, 0, 0, 0, 0, 0, 0, 0};
  return val[location]++;
  */
	// updated - might be slower...
	static unsigned int val[] = {0, 0, 0, 0, 0, 0, 0, 0};
	if( threads ) {
		pthread_mutex_lock( &mainMutex[location] );
		unsigned int tmp = val[location]++;
		pthread_mutex_unlock( &mainMutex[location] );
		return tmp;
	}
	return val[location]++;
}
float min( float left, float right ) {
  return left < right ? left : right;
}
float max( float left, float right ) {
  return left > right ? left : right;
}

float invsqrt( float value ) {
  return 1. / sqrt(value);
}

// Debug
void profile( int prof_id ) {}
int loadl1( int base, int offset = 0 ) {return 0;}
void trax_printi(int value) { printf("Value is %d.\n", value);}
void trax_printf( float value ) { printf("Value is %f.\n", value);}

int start_framebuffer;

// 2^25
//#define MEMORY_SIZE 33554432
#define MEMORY_SIZE 67108864		// 2^26
#ifndef VIEWFILE
#define VIEWFILE "views/cornell_obj.view"
#endif
#ifndef MODELFILE
#define MODELFILE "test_models/cornell/CornellBox.obj"
#endif
#ifndef LIGHTFILE
#define LIGHTFILE "lights/cornell.light"
#endif

void trax_setup() {
  // set up memory
  trax_memory_pointer = new FourByte[MEMORY_SIZE]; 

  BVH* bvh;
  int grid_dimensions = -1;
  //  char* view_file = VIEWFILE;
  Camera* camera = ReadViewfile::LoadFile(VIEWFILE);
  //  char* model_file = MODELFILE;
  int start_wq, start_scene, start_camera, start_bg_color, start_light, end_memory;
  int start_matls, start_permutation;
  //  char* light_file = LIGHTFILE;
  float *light_pos = new float[3];
  ReadLightfile::LoadFile(LIGHTFILE, light_pos);
  int tile_width = 16;
  int tile_height = 16;
  int ray_depth = 1;
  int num_samples = 1;
  int num_thread_procs = 1; // not sure why this is needed
  int num_cores = 1; // this either
  float epsilon = 1e-4;
  bool duplicate_bvh = false;

  LoadMemory(trax_memory_pointer, bvh, MEMORY_SIZE, image_width, image_height,	
	     grid_dimensions,
	     camera, MODELFILE,
	     start_wq, start_framebuffer, start_scene,
	     start_matls,
	     start_camera, start_bg_color, start_light, end_memory,
	     light_pos, start_permutation, tile_width, tile_height,
	     ray_depth, num_samples, num_thread_procs * num_cores, epsilon, duplicate_bvh);
}

void trax_cleanup(){
  
	// since I haven't figured out how to dump stuff out in windows using pipes,
	// I'll just resort to saving out PPM files
#if defined(WIN32) && TRAX==0
	FILE* output = fopen( "out.ppm" , "wb" );
	if (!output)
		perror("Failed to open out.ppm");
    
	fprintf(output, "P6\n%d %d\n%d\n", image_width, image_height, 255);
	for (int j = image_height - 1; j >= 0; j--) {
		for (int i = 0; i < image_width; i++) {
			int index = start_framebuffer + 3 * (j * image_width + i);
			float rgb[3];
			// for gradient/colors we have the result
			rgb[0] = trax_memory_pointer[index + 0].fvalue;
			rgb[1] = trax_memory_pointer[index + 1].fvalue;
			rgb[2] = trax_memory_pointer[index + 2].fvalue;
			static unsigned char color[3];
			color[0] = (char)(unsigned int)(rgb[0] * 255);
			color[1] = (char)(unsigned int)(rgb[1] * 255);
			color[2] = (char)(unsigned int)(rgb[2] * 255);
			fwrite( color, 1, 3, output );
		}
	}
	fclose( output );
#else
  // output the image
  char command_buf[512];
  sprintf(command_buf, "convert PPM:- out.png");
  FILE* output = popen(command_buf, "w");
  if (!output)
    perror("Failed to open out.png");
	
  fprintf(output, "P6\n%d %d\n%d\n", image_width, image_height, 255);
  for (int j = image_height - 1; j >= 0; j--) {
    for (int i = 0; i < image_width; i++) {
      int index = start_framebuffer + 3 * (j * image_width + i);
	    
      float rgb[3];
      // for gradient/colors we have the result
      rgb[0] = trax_memory_pointer[index + 0].fvalue;
      rgb[1] = trax_memory_pointer[index + 1].fvalue;
      rgb[2] = trax_memory_pointer[index + 2].fvalue;
      fprintf(output, "%c%c%c",
	      (char)(int)(rgb[0] * 255),
	      (char)(int)(rgb[1] * 255),
	      (char)(int)(rgb[2] * 255));
    }
  }
  pclose(output);
#endif
}


/// creates a bunch of threads to run rendering function
bool trax_start_render_threads( void* (*renderFunc)(void*), const int &numThreads ) {

	// just in case
	if( threads )
		return false;

	threads = new pthread_t[numThreads];

	// init mutexes
	for( unsigned int m=0; m<numAtomicInc; ++m ) {
		if( pthread_mutex_init( &mainMutex[m], NULL ) ) {
			printf( "Unable to initialize mutex %d\n", m );
			return false;
		}
	}

	// start up the rendering...
	for( int i=0; i<numThreads; ++i ) {
		if( pthread_create( &threads[i], NULL, renderFunc, 0 ) ) {
			printf( "Could not create thread %d\n", i );
			return false;
		}
	}

	// make sure all the threads exit upon completion
	for( int i=0; i<numThreads; ++i ) {
		if( pthread_join( threads[i], NULL ) ) {
			printf( "Could not join thread %d\n", i );
			return false;
		}
	}

	// clear out the memory
	delete[] threads;
	for( unsigned int m=0; m<numAtomicInc; ++m )
		pthread_mutex_destroy( &mainMutex[m] );
	return true;
}